# http polling interface http://www.jabber.org/jeps/jep-0025.html
# $Id: httpp.py,v 1.32 2005/11/07 14:03:03 tofu Exp $
from twisted.python import components
from twisted.web import server, resource
from twisted.internet import reactor, protocol, defer

try:
    from twisted.words.xish import domish, utility
except:
    from twisted.xish import domish, utility

from zope.interface import Interface, implements

from xmpp import component

import sha, base64, sys
import session, error
import types
import punjab

import traceback
from xmpp import client

Error = error.Error

class BadRequest(Error):
    stanza_error = 'bad-request'
    msg = 'bad-request'
    def __init__(self,msg='bad-request'):
        Error.__init__(self,msg)
    
class NotFound(Error):
    def __init__(self,msg='not-found'):
        Error.__init__(self,msg)
        

    
class InternalServerError(Error):
    def __init__(self,msg='internal-server-error'):
        Error.__init__(self,msg)


class RemoteConnectionFailed(Error):
    def __init__(self,msg='remote-connection-failed'):
        Error.__init__(self,msg)



# make httpp body class, similar to xmlrpclib
# 
class HttppParse:
    def __init__(self):
        # Setup the parser
        self.stream = domish.elementStream()
        self.stream.DocumentStartEvent = self.onDocumentStart
        self.stream.ElementEvent = self.onElement
        self.stream.DocumentEndEvent = self.onDocumentEnd
        
        self.xmpp_elements = []
        
        self.req        = {}
        self.req['rid'] = '0'

    def parse(self,buf):
        if buf.find(',')!=-1:
            arr = buf.split(',',1)
            req = arr.pop(0)
            self.xml = arr[0]
            
            if req.find(';')!=-1:
                rarr = req.split(';',1)
                self.req['rid'] = rarr.pop(0)
                self.req['key'] = q.join(rarr)
                if self.req['key'].find(';')!=-1:
                    self.req['key'],self.req['newkey'] = self.req['key'].split(';',1)
                if self.req['rid']=='0':
                    self.req['newkey'] = self.req['key']
            else:
                self.req['rid'] = req
                
            if self.xml == '</stream:stream>':
                return self.req,0
        
            # TODO - fix this hack!
            if self.req['rid'] != '0':
                self.xml = '<tmp xmlns="jabber:client">'+self.xml+'</tmp>'       
        
            self.stream.parse(self.xml)
            if self.req['rid'] != '0':
                self.xmpp_elements.pop(0)
            
            return self.req,self.xmpp_elements
        else:
            return 0,0
    
    def serialize(self, obj):
        if isinstance(obj, domish.Element):
            obj = obj.toXml()
        return obj

    def onDocumentStart(self, rootelem):
        if rootelem!='tmp':
            self.xmpp_elements.append(rootelem)
        

    def onElement(self, element):
        if isinstance(element, domish.Element):
            self.xmpp_elements.append(element)
        else:
            pass


    def onDocumentEnd(self):
        pass


class IHttppService(Interface):

    def __init__(self, verbose):
        """ """

    def startSession(self, body):
        """ Start a punjab jabber session """

    def endSession(self, body):
        """ end a punjab jabber session """

    def parseReq(self, body):
        """ parse an body element """

    def inSession(self,sid):
        """ """
        

        
class IHttppFactory(Interface):

    def startSession(self):
        """ Start a punjab jabber session """

    def endSession(self, sid):
        """ end a punjab jabber session """

    def parseReq(self, body):
        """ parse an body element """

    def buildProtocol(self, addr):
        """Return a protocol """



    
class Httpp(resource.Resource):
    isLeaf = True
    def __init__(self,service,v=0):
        """Initialize.
        """
        resource.Resource.__init__(self)
        self.service  = service
        self.children = {}
        self.counter  = 0
        self.client   = 0
        self.v        = v
        
    def render_GET(self, request):
        return "<html><body><a href='http://www.xmpp.org/extensions/xep-0025.html'>XEP-0025</a> - HTTP Polling</body></html>"

    def render_POST(self, request):
        request.content.seek(0, 0)       
        hp = HttppParse()
        try:
            req,xmpp_elements = hp.parse(request.content.read())            
            sid = request.getCookie            
        except:
            self.send_http_error(400,request) 
            return server.NOT_DONE_YET
        else:
            if isinstance(req,int):
                self.send_http_error(500,request) 
                return server.NOT_DONE_YET
            if self.service.inSession(req):                
                s, d = self.service.parseReq(req, xmpp_elements)
                d.addCallback(self.return_httpp, s, request)
            else:
                if request.args.has_key('server'):
                    req['to'] = request.args['server'][0]
                s, d = self.service.startSession(req, xmpp_elements)
                d.addCallback(self.return_session, s, request)

            d.addErrback(self.return_error, request)
                
        return server.NOT_DONE_YET
        

    def return_session(self, data, session, request):
        # create stream
        sh = "<stream:stream xmlns='%s' xmlns:stream='http://etherx.jabber.org/streams' id='%s' from='%s'>" % \
                 ('jabber:client', session.authid.encode(session.charset), session.to.encode(session.charset))

        while len(data) > 0:
            data.pop(0).toXml()
            
        request.addCookie('ID',session.sid)
        self.return_request(request,sh)


    def return_error(self, error,  request):
        print error
        self.send_http_error(500,request,'internal-server-error','error',error)
        
    def return_httpp(self, data, session, request):
        xml = None
        ret = ''
        request.addCookie('ID',session.sid)  
        
        while len(data) > 0:
            ret = ret + data.pop(0).toXml()
            
        self.return_request(request,ret, session.charset)

    def return_request(self,request,o, charset = 'utf-8'):
        request.setResponseCode(200)
        
        out = o.encode('utf-8','replace')
        # out = o
            
        request.setHeader('content-type','text/xml')
        request.setHeader('content-length', str(len(out)))
        request.write(out)
        request.finish()
            
    def send_http_error(self,code,request, condition = 'undefined-condition', type = 'terminate', data = ''):
        request.setResponseCode(code)
        request.setHeader("content-length", "0")
        request.finish()



components.registerAdapter(Httpp, IHttppService, resource.IResource)


class HttppService(punjab.Service):

    implements(IHttppService)

    def __init__(self, verbose):
        self.v  = verbose
        self.sessions = {}
        self.counter  = 0
        
    def startSession(self, req, xmpp_elements):
        """ Start a punjab jabber session """        
        try:
            if req['rid']=='0':
                if len(xmpp_elements)>0:
                    try:
                        req['to'] = xmpp_elements[0]['to']
                    except:
                        traceback.print_exc()
                        raise InternalServerError
                
                    
            else:
                traceback.print_exc()
                raise InternalServerError
            req['wait'] = 0
            req['hold'] = 0

            if not req.has_key('to'):
                raise InternalServerError
            
            return session.make_session(self,req)
        except:
            traceback.print_exc()
            raise InternalServerError

    def endSession(self, session):
        """ end a punjab jabber session """
        d = session.terminate()
        return d

    def parseReq(self, req, xmpp_elements):
        try:    
            s = self.sessions[req['rid']]
            s.touch() # any connection should be a renew on wait
            if xmpp_elements == 0:
                d = s.terminate()
                return s, d
            try:
                if req.has_key('key'):
                    if s.key is not None:
                        nk = sha.new(req['key'])
                        key = base64.encodestring(nk.digest())
                        next_key = req['key']
                        key = key.rstrip()
                        if key == s.key:
                            s.key = next_key
                            if req.has_key('newkey'):
                                newkey = req['newkey']
                                s.key = newkey
                        else:
                            raise NotFound
                    else:
                        traceback.print_exc()
                        raise NotFound
            except:
                traceback.print_exc()
                raise NotFound
            
            return s, self._parse(s, req, xmpp_elements)
                        
        except:
            traceback.print_exc()
            raise InternalServerError
            
    def _parse(self, session, body_tag, xmpp_elements):
        d = None
        # increment the request counter
        session.rid  = session.rid + 1
        iq_send = False
        try:
            # send all the elements
            for el in xmpp_elements:
                if el.name == 'iq' and  el.firstChildElement().name == 'query' \
                   and (el.firstChildElement().uri == 'jabber:iq:auth' or \
                        ( el.firstChildElement().hasAttribute('xmlns') and  el.firstChildElement()['xmlns'] == 'jabber:iq:auth')):
                    
                    # NOTE - temp hack from iq:auth
                    #        need to fix this, jwchat expects an iq return
                    d = defer.Deferred()
                    iq = client.IQ(session.xmlstream,el['type'])
                    iq.attributes = el.attributes
                    if not iq.hasAttribute('id'):
                        iq.addUniqueId()
                    iq.children   = el.children
                    iq.addCallback(self._returnIq, session, d)
                    iq.send()
                    iq_send = True
                else:
                    session.sendRawXml(el)
                
            if body_tag.has_key('type') and \
               body_tag['type'] == 'terminate':
                d = session.terminate()
            elif iq_send == False:
                # normal request
                d = session.poll(d)

            return d
        except:
            traceback.print_exc()
            raise InternalServerError
    

    def _returnIq(self, session, d, iq):
        session.elems.append(iq)
        return session.poll(d)
        
    def inSession(self,req):
        """ """
        if req.has_key('rid'):
            if self.sessions.has_key(req['rid']):        
                return True
        return False
                
# This is the jabberd commponent connect classes and code


class HttppComponentService(component.Service):
    def componentConnected(self, xmlstream):
        self.xmlstream = xmlstream
        xmlstream.addObserver(VERSION, self.onVersion, 1)
        

    def onVersion(self, iq):
        iq.swapAttributeValues("to", "from")
        iq["type"] = "result"
        name = iq.addElement("name", None, 'Punjab')
        version = iq.addElement("version", None, __version__)
        self.send(iq)
        iq.handled = True

