"""
 session stuff for jabber connections

"""
from twisted.internet import defer,  reactor
from twisted.python import log
from twisted.web import server
from twisted.names.srvconnect import SRVConnector

from twisted.words.protocols.jabber import jid
try:
    from twisted.words.xish import domish, xmlstream
except:
    from twisted.xish import domish, xmlstream


import  md5, random, base64, types, binascii, os, traceback
import jabber
from xmpp import disco, client
from xmpp.ns import *
import time
import error


def _gen_nonce():
    """
    Create a SASL nonce
    """
    return md5.new("%s:%s:%s" % (str(random.random()) , str(time.gmtime()), str(os.getpid()))).hexdigest()



class XMPPClientConnector(SRVConnector):
    """
    A jabber connection to find srv records for xmpp client connections.
    """
    def __init__(self, client_reactor, domain, factory):
        """ Init """
        SRVConnector.__init__(self, client_reactor, 'xmpp-client', domain, factory)


    def pickServer(self):
        """
        Pick a server and port to make the connection.
        """
        host, port = SRVConnector.pickServer(self)
        
        if not self.servers and not self.orderedServers:
            # no SRV record, fall back..
            port = 5222
        if port == 5223 and xmlstream.ssl:
            context = xmlstream.ssl.ClientContextFactory()
            context.method = xmlstream.ssl.SSL.SSLv23_METHOD
            
            self.connectFunc = 'connectSSL'
            self.connectFuncArgs = (context)
        return host, port

def make_session(pint, attrs):
    """
    pint  - punjab session interface class
    attrs - attributes sent from the body tag
    """    

    # this may need some work, idea, code taken from twisted.web.server
    pint.counter = pint.counter + 1
    sid  = md5.new("%s_%s" % (str(random.random()) , str(pint.counter))).hexdigest()
    
    s    = Session(pint, sid, attrs)

    s.addBootstrap(xmlstream.STREAM_START_EVENT, s.streamStart)
    s.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, s.connectEvent)
    s.addBootstrap(xmlstream.STREAM_ERROR_EVENT, s.streamError)
    s.addBootstrap(xmlstream.STREAM_END_EVENT, s.connectError)    

    if attrs.has_key('inactivity'):
        s.inactivity = attrs['inactivity']
    else:
        s.inactivity = 900 # 15 mins
        
    s.secure = 0
    
    if attrs.has_key('secure') and attrs['secure'] == 'true':
        s.secure = 1
        s.authenticator.useTls = 0
        
        
    else:
        s.authenticator.useTls = 1
        # reactor.connectTCP(s.hostname, s.port, s)
    log.msg('================================== %s connect ==================================' % (str(time.time()),))
    connector = XMPPClientConnector(reactor, s.hostname, s)
    connector.connect()
    # timeout
    reactor.callLater(s.inactivity, s.checkExpired)

    pint.sessions[sid] = s
    
    return s, s.waiting_requests[0].deferred
    

class WaitingRequest(object):
    """A helper object for managing waiting requests."""

    def __init__(self, deferred, delayedcall, timeout = 30, startup = False, rid = None):
        """ """
        self.deferred = deferred
        self.delayedcall = delayedcall
        self.startup = startup
        self.timeout = timeout
        self.wait_start = time.time()
        self.rid = rid
        
    def doCallback(self, data):
        """ """
        self.deferred.callback(data)

    def doErrback(self, data):
        """ """
        self.deferred.errback(data)


class Session(jabber.JabberClientFactory, server.Session):
    """ Jabber Client Session class for client XMPP connections. """
    def __init__(self, pint, sid, attrs):
        """
        Initialize the session
        """
        if attrs.has_key('charset'):
            self.charset = str(attrs['charset'])
        else:
            self.charset = 'utf-8'
        
        self.to    = attrs['to']
        self.port  = 5222
        self.inactivity = 900
        if self.to != '' and self.to.find(":") != -1:
            # Check if port is in the 'to' string
            to, port = self.to.split(':')
            
            if port:
                self.to   = to
                self.port = int(port)
            else:
                self.port = 5222
        
        jabber.JabberClientFactory.__init__(self, self.to, pint.v)
        server.Session.__init__(self, pint, sid)
        self.pint  = pint
        self.sid   = sid
        self.attrs = attrs
        self.s     = None
        self.iqs   = []
        self.msgs  = []
        self.prs   = []
        self.elems = []
        rid        = int(attrs['rid'])

        self.waiting_requests = []

        self.xmpp_node = ''       

        self.chat_p = 0
        self.success = 0
        self.bind = 0
        self.secure = 0
        self.mechanisms = []
        self.xmlstream  = None
        self.features   = None
        self.session    = None
        self.passwd     = None
        self.cache_data = {}
        
        if attrs.has_key('version'):
            self.version = attrs['version']
        else:
            self.version = 0.0
        
        if attrs.has_key('newkey'):
            #newkey   = base64.decodestring(attrs['newkey'])
            newkey   = attrs['newkey']
            self.key = newkey
        
        if attrs.has_key('wait'):
            self.wait  = int(attrs['wait'])            
        else:
            self.wait  = 0

        if attrs.has_key('hold'):
            self.hold  = int(attrs['hold'])
        else:
            self.hold  = 0

        if attrs.has_key('window'):
            self.window  = int(attrs['window'])
        else:
            self.window  = self.hold + 2

        if attrs.has_key('polling'):
            self.polling  = int(attrs['polling'])
        else:
            self.polling  = 0
           
        if attrs.has_key('port'):
            self.port = int(attrs['port'])

        if attrs.has_key('hostname'):
            self.hostname = attrs['hostname']
        else:
            self.hostname = self.to
            
        if attrs.has_key('route'):
            if attrs['route'].startswith("xmpp:"):
                self.route = attrs['route'][5:]
                if self.route.startswith("//"):
                    self.route = self.route[2:]

                # route format change, see http://www.xmpp.org/extensions/xep-0124.html#session-request
                rhostname, rport = self.route.split(":")
                self.port = int(rport)
                self.hostname = rhostname
                                                         
                self.resource = ''
                
                
            else:
                raise error.Error('internal-server-error')

            
        self.authid      = 0    
        self.rid         = rid + 1
        self.connected   = 0 # number of clients connected on this session 
        self.group_chats = {}
        self.notifyOnExpire(self.onExpire)
        if pint.v:
            log.msg('Session Created : %s %s' % (str(self.sid),str(time.time()), ))
        
        # create the first waiting request
        d = defer.Deferred()
        timeout = 30
        self.waiting_requests.append(
            WaitingRequest(d,
                           self._startup_timeout,
                           startup = True,
                           timeout = timeout,
                           rid = self.rid - 1))
        

    def rawDataIn(self, buf):
        """ Log incoming data on the xmlstream """
        log.msg("SID: %s => RECV: %s" % (self.sid, unicode(buf, 'utf-8').encode('ascii', 'replace')))

    def rawDataOut(self, buf):
        """ Log outgoing data on the xmlstream """
        log.msg("SID: %s => SEND: %s" % (self.sid, unicode(buf, 'utf-8').encode('ascii', 'replace')))
        

    def onExpire(self):
        """ When the session expires call this. """
        if 'onExpire' in dir(self.pint):
            self.pint.onExpire(self.sid)
        log.msg(self.sid)
        log.msg(self.rid)
        log.msg('SESSION -> We have expired')
        self.disconnect()
    
    def terminate(self):
        """Terminates the session."""
        self.wait = 0
        log.msg('SESSION -> Terminate')
        # if there are any elements hanging around and waiting
        # requests, send those off
        while len(self.elems) > 0 and len(self.waiting_requests) > 0:
            data = self.elems
            self.elems = []
            
            wr = self.waiting_requests.pop(0)
            wr.doCallback(data)

        # if there are any waiting requests, send them back blank
        while len(self.waiting_requests) > 0:
            wr = self.waiting_requests.pop(0)
            wr.doCallback([])
        try:
            self.expire()
        except:
            self.onExpire()
        
                
        return defer.succeed(self.elems)

    def poll(self, d = None):
        """Handles the responses to requests.

        This function is called for every request except session setup
        and session termination.  It handles the reply portion of the
        request by returning a deferred which will get called back
        when there is data or when the wait timeout expires.
        """
        
        # queue this request
        if d is None:
            d = defer.Deferred()
        if self.pint.error:
            d.addErrback(self.pint.error)    
        
        self.waiting_requests.append(
            WaitingRequest(d,
                           self._pollTimeout,
                           timeout = self.wait,
                           rid = self.rid - 1))
            
        
        # check if there is any data to send back to a request
        if len(self.elems) > 0:
            data = self.elems
            # pop off a request and send and reply
            if len(self.waiting_requests)>0:
                self.elems = []                            
                wr = self.waiting_requests.pop(0)
                wr.doCallback(data)
                self._cacheData(wr.rid, data)
                
        # make sure we aren't queueing too many requests
        while len(self.waiting_requests) > self.hold:
            if len(self.elems) > 0:
                data = self.elems
            else:
                data = []
            wr = self.waiting_requests.pop(0)
            wr.doCallback(data)
            self._cacheData(wr.rid, data)
            
        return d

    def _pollTimeout(self, d):
        """Handle request timeouts.

        Since the timeout function is called, we must return an empty
        reply as there is no data to send back.
        """
        
        # find the request that timed out and reply
        pop_eye = []
        for i in range(len(self.waiting_requests)):
            if self.waiting_requests[i].deferred == d:
                wr = self.waiting_requests[i]
                pop_eye.append(i)
                self.touch()
                wr.doCallback([])
        for i in pop_eye:
            wr = self.waiting_requests.pop(i)

    def _pollForId(self, d):
        if self.xmlstream.sid:
            self.authid = self.xmlstream.sid
        self._pollTimeout(d)
        

    # start login - 
    def startLogin(self, jabber_id, passwd):
        self.touch()
        # create the login waiting request
        d = defer.Deferred()	
            
        d.addErrback(self.pint.error)    
        
        self.waiting_requests.append(
        	    WaitingRequest(d,
                           	None,
                                   rid = self.rid - 1,
                                    ))
                        
        self.jid    = jid.JID(jabber_id)
        
        self.passwd = passwd        
        # send login with callback

        if  self.authenticator.version == 0.0:
            self.authenticator.sendAuth(self.jid, passwd, self._authQueryResultEvent, self._authErrEvent)
        else:
            ms = None
            # DO SASL stuff. TODO - move in authenticator or move in twisted.words.protocol.jabber.client
            for mech in self.mechanisms:
                ms = str(mech)
                if ms == 'DIGEST-MD5':
                    break
                if ms == 'PLAIN':
                    break
            if ms is None:
                self.authenticator.sendAuth(self.jid, passwd, self._authQueryResultEvent, self._authErrEvent)
                return d
            
            auth = domish.Element((client.SASL_XMLNS, "auth"), client.SASL_XMLNS,{'mechanism' : ms})

            auth['mechanism'] = ms
            if ms == 'DIGEST-MD5':
                self.xmlstream.addOnetimeObserver("/challenge", self._saslStep1)
            elif ms == 'PLAIN':
                # TODO add authzid

                auth_str = ""
                auth_str = auth_str + "\000"
                auth_str = auth_str + self.jid.user.encode(self.charset)
                auth_str = auth_str + "\000"
                auth_str = auth_str + self.passwd.encode(self.charset)
                auth.addContent(binascii.b2a_base64(auth_str))
        
                self.xmlstream.addOnetimeObserver("/success", self._saslSuccess, d)
                
            self.xmlstream.addObserver("/failure", self._saslError, d)
            self.xmlstream.send(auth)
        return d


    def _authQueryResultEvent(self, xs):
        if len(self.waiting_requests)>0:		
		wr = self.waiting_requests.pop(0)
		d = wr.deferred
	else:
		d = None
                
        if xs is None:
            d.errback(self)
            return
        self.authenticator.authQueryResultEvent(xs, self._authResultEvent, d)

        
    def _authResultEvent(self, iq, d = None):

	if d:
            if isinstance(iq, domish.Element):
                if iq["type"] == "result":
                    d.callback(self)
                else:
                    d.errback(self)
            else:
                d.errback(self)
            

    def _authErrEvent(self, err, d = None):
        log.err(err.printTraceback())
        d.errback(self)
        
    def connectEvent(self, xs):

        self.version =  self.authenticator.version
        self.xmlstream = xs
        if self.pint.v != 0:
            # add logging for verbose output
            self.xmlstream.rawDataInFn = self.rawDataIn
            self.xmlstream.rawDataOutFn = self.rawDataOut
        log.msg(self.version)
        if self.version == '1.0':
            self.xmlstream.addObserver("/features[@xmlns='http://etherx.jabber.org/streams']", self.featuresHandler)

            
    def streamStart(self, xs):
        """
        A xmpp stream has started
        """

        log.msg(self.xmlstream.sid)
        # self.xmlstream.prefixes = XMPP_PREFIXES
        # This is done to fix the stream id problem, I should submit a bug to twisted bugs
        
        try:
            
            self.authid    = self.xmlstream.sid
            
            if not self.attrs.has_key('no_events'):

                from httpb import HttpbService
                from httpp import HttppService
                # TODO - this needs fixing, we need a way to set the different observers
                #        from the Service (pint)
                #        Also, we need a way to handle all elements
                if isinstance(self.pint, HttpbService):
                    self.success = 1
                    self.xmlstream.addOnetimeObserver("/auth", self.stanzaHandler)
                    self.xmlstream.addOnetimeObserver("/response", self.stanzaHandler)
                    self.xmlstream.addOnetimeObserver("/success", self._saslSuccess)
                    self.xmlstream.addOnetimeObserver("/failure", self._saslError)                    

                    self.xmlstream.addObserver("/iq/bind", self.bindHandler)
                    self.xmlstream.addObserver("/bind", self.stanzaHandler)
                    
                    self.xmlstream.addObserver("/challenge", self.stanzaHandler)
                    self.xmlstream.addObserver("/message",  self.stanzaHandler)
                    self.xmlstream.addObserver("/iq",  self.stanzaHandler)
                    self.xmlstream.addObserver("/presence",  self.stanzaHandler)
                    # TODO - we should do something like this
                    # self.xmlstream.addObserver("/*",  self.stanzaHandler)
                elif isinstance(self.pint, HttppService):
                    self.success = 1
                    self.xmlstream.addObserver("/message",  self.stanzaHandler)
                    self.xmlstream.addObserver("/iq",  self.stanzaHandler)
                    self.xmlstream.addObserver("/presence",  self.stanzaHandler)
                else:
                    self.xmlstream.addObserver("/message", self.msgHandler, -1)
                    self.xmlstream.addObserver("/presence", self.preHandler, -1)
            
            if self.version != '1.0':
                wr = self.waiting_requests.pop(0)
                if self.xmlstream.sid is None:
                    # poll for the id
                    reactor.callLater(0, self._pollForId, wr)
                else:
                    wr.doCallback([])
        except:
            log.err(traceback.print_exc())
            wr = self.waiting_requests.pop(0)
            
            wr.doErrback(error.Error("remote-connection-failed"))
                
            self.disconnect()
            
    # This is here just for old stuff
    def featuresHandler(self, f):
        f.prefixes = XMPP_PREFIXES
        self.bind    = 0
        self.session = 0
        #check for tls
        self.f = {}
        for feature in f.elements():
            self.f[(feature.uri, feature.name)] = feature
        
        starttls = (client.TLS_XMLNS, 'starttls') in self.f
        # TODO - this check needs to be cleaned up
        useTls         = getattr(self.xmlstream, 'useTls' , 0)
        tlsEstablished = getattr(self.xmlstream, 'tlsEstablished', False)
        
        if tlsEstablished or useTls == 0 or \
           (useTls == 1 and not starttls):
            if tlsEstablished:
                self.secure = 1           
            # pass
        elif not tlsEstablished and useTls in (1, 2) and starttls:
            return
        else:
            self.secure = 0

        self.features = f
        if self.authid is None:
            self.authid = self.xmlstream.sid
        if self.success == 1:
            # TODO - create the abilitly to change the featuresHandler per service
            from httpb import HttpbService
            from httpp import HttppService        
            if isinstance(self.pint, HttpbService) or isinstance(self.pint, HttppService):
                
                # If we get tls, then we should start tls and then return 
                if len(self.waiting_requests) > 0:
                    wr = self.waiting_requests.pop(0)
                    wr.doCallback([f])
                else:
                    self.elems.append(f)
                return
            
            for f in f.elements():
                if f.name == "bind":
                    self.bind = 1
                if f.name == "session":
                    self.session = 1
                    
            if self.bind == 1:
                iq = client.IQ(self.xmlstream, "set")
                iq.addElement((client.BIND_XMLNS, "bind"))
                
                iq.bind.addElement("resource", content = self.jid.resource)
                iq.addCallback(self._bindResultEvent)
                iq.send()
                
            return
        else:
            self.m = f.mechanisms

            if len(self.waiting_requests)>0:
                wr = self.waiting_requests.pop(0)
                try:
                    if self.m.uri == client.SASL_XMLNS:
                        for mech in self.m.elements():
                            self.mechanisms.append(str(mech))
                        # TODO - move to twisted.words.protocols.jabber 
                
                    else:
                        if wr:
                            wr.doErrback(error.Error('remote-stream-error'))
                except:
                    pass
            
                if wr:
                    wr.doCallback([])

    def bindHandler(self, stz):
        """bind debugger for punjab, this is temporary! """
        
        log.msg('======================== bindHandler ====================')
        try:
            log.msg('BIND: %s %s' % (str(self.sid), str(stz.bind.jid)))
        except:
            log.err()
        
        
    def stanzaHandler(self, stz):
        """generic stanza handler for httpbind and httppoll"""
        stz.prefixes = XMPP_PREFIXES
        if self.waiting_requests and len(self.waiting_requests) > 0:
            # if there are any waiting requests, give them all the
            # data so far, plus this new data

            wr = self.waiting_requests.pop(0)

            data = self.elems + [stz]
            self.elems = []

            wr.doCallback(data)
            self._cacheData(wr.rid, data)
        else:
            # since there are no waiting requests, just queue the data
            self.elems.append(stz)

    def _startup_timeout(self, d):
        # this can be called if connection failed, or if we connected
        # but never got a stream features before the timeout
        log.msg('================================== %s %s startup timeout ==================================' % (str(self.sid), str(time.time()),))
        for i in range(len(self.waiting_requests)):
            if self.waiting_requests[i].deferred == d:
                wr = self.waiting_requests.pop(i)
                # check if we really failed or not
                if self.authid:
                    if len(self.elems) > 0:
                        data = self.elems
                    else:
                        data = []
                    wr.doCallback(data)
                else:
                    wr.doErrback(error.Error("remote-connection-failed"))
                    
    
    def addMessageObserver(self, event, cb, priority = 0):
        self.xmlstream.removeObserver("/message",  self.msgHandler)
        self.xmlstream.addObserver(event,  cb, priority)
        if len(self.msgs)>0:
            msgs = self.msgs
            self.msgs = []
            return msgs
        else:
            return None
    
    def addPresenceObserver(self, event, cb, priority = 0):
        self.xmlstream.removeObserver("/presence",  self.preHandler)
        self.xmlstream.addObserver(event,  cb, priority )
        
        if len(self.prs)>0:
            prs = self.prs
            self.prs = []
            return prs
        else:
            return None
    
    def streamError(self, streamerror):
        """called when we get a stream:error stanza"""
        
        
        e = error.Error('remote-stream-error')
        e.error_stanza = 'remote-stream-error'
        e.children = []
        try: # a workaround for a bug in twisted.words.protocols.jabber.error
            err_elem = streamerror.value.getElement()
            err_elem.toXml()
            e.children.append(err_elem)
        except:
            pass
        
        if len(self.waiting_requests) > 0:
            wr = self.waiting_requests.pop(0)            
            wr.doErrback(e)
        if self.pint and self.pint.sessions.has_key(self.sid):
            try:
                self.expire()
            except:
                self.onExpire()

    def connectError(self, xs):
        """called when we get disconnected"""
        
        # FIXME: we should really only send the error event back if
        # attempts to reconnect fail.  There's no reason temporary
        # connection failures should be exposed upstream
        #if self.pint.v != 0:
        log.msg('connect ERROR')
        try:
            log.msg(xs)
            log.msg(xs.toXml())
        except:
            pass
            
        if self.waiting_requests:
            log.msg(str(len(self.waiting_requests)))
            log.msg(xs)
            if len(self.waiting_requests) > 0:
                wr = self.waiting_requests.pop(0)
                wr.doErrback(error.Error('remote-connection-failed'))

        if self.pint and self.pint.sessions.has_key(self.sid):
            try:
                self.expire()
            except:
                self.onExpire()


    def sendRawXml(self, obj):
        """
        Send a raw xml string, not a domish.Element
        """
        self.touch()
        self._send(obj)
    
    def sendXml(self, obj):
        """
        Serialize the element and send it over the wire.
        """
        self.touch()
        d = defer.Deferred()
        
        if obj.name == 'iq':
            iq = client.IQ(self.xmlstream, obj['type'])
            iq.attributes = obj.attributes
            if not iq.hasAttribute('id'):
                iq.addUniqueId()
            iq.children   = obj.children
            iq.addCallback(self._returnIq, d)
            iq.send()
        elif obj.name == 'response':
            mech = None
                       
            if str(obj) == '':
                self.xmlstream.addOnetimeObserver("/success", self._saslSuccess, d)
            else:
                self.xmlstream.removeObserver('/challenge',  self._saslStep1, d)
                self.xmlstream.addObserver('/challenge',  self._saslStep2, d)
            self.xmlstream.send(obj)
            
        elif obj.name == 'auth':
            mech = None
            
            for m in self.mechanisms:
                if obj['mechanism'] == m:
                    mech = m
            if mech and mech != "PLAIN" and mech != "ANONYMOUS":    
                self.xmlstream.addObserver('/challenge',  self._saslStep1, d)
                
                self.xmlstream.send(obj)
            elif mech == "PLAIN" or mech == "ANONYMOUS":
        
                self.xmlstream.addOnetimeObserver("/success", self._saslSuccess, d)
                log.msg('SASL PLAIN ANONYMOUS\n')
                self.xmlstream.addObserver("/failure", self._saslError, d)
                self.xmlstream.send(obj)
            else:
                return -1
        else:
            self._send(obj)
        return d

    
    def _send(self, xml):
        """
        Send valid data over the xmlstream
        """
        if self.xmlstream: # FIXME this happens on an expired session and the post has something to send
            self.xmlstream.send(xml)

    def _removeObservers(self, typ = ''):
        if typ == 'event':
            observers = self.xmlstream._eventObservers
        else:
            observers = self.xmlstream._xpathObservers
        emptyLists = []
        for priority, priorityObservers in observers.iteritems():
            for query, callbacklist in priorityObservers.iteritems():
                callbacklist.callbacks = []
                emptyLists.append((priority, query))

        for priority, query in emptyLists:
            del observers[priority][query]
    
    def disconnect(self):
        """
        Disconnect from the xmpp server.
        """
        if not getattr(self, 'xmlstream',None):
            return
        
        if self.xmlstream:
            #sh = "<presence type='unavailable' xmlns='jabber:client'/>"
            sh = "</stream:stream>"
            self.xmlstream.send(sh)
            
        self.stopTrying()
        if self.xmlstream:
            self.xmlstream.transport.loseConnection()
            # we should remove observers
            #self._removeObservers('event')
            #self._removeObservers()

            del self.xmlstream
        self.connected = 0
        self.pint      = None
        self.elems     = []
        
        if self.waiting_requests:
            for i in range(len(self.waiting_requests)):
                wr = self.waiting_requests.pop(i)
                wr.doCallback()
            del self.waiting_requests
        self.mechanisms = None
        self.features   = None
        
    def sendPresence(self, to, type, show = '', status = '', priority = 0):
        """
        Send a presence xmpp element.
        """
        self.touch()
        try:
            presence = domish.Element(("jabber:client","presence"))
            if to != '':
                presence['to'] = to
            if type != '':
                presence['type'] = type
            if show != '':
                presence.addElement('show', content = show)
            
            if status != '':
                presence.addElement('status', content = status)

            if priority != 0:
                presence.addElement('priority', content = priority)
        except:
            return 0
        self.xmlstream.send(presence)
        
        return 1

    def sendMessage(self, to, subject, body, typ, m_id):
        """
        Send a message xmpp element
        """
        self.touch()
        message = domish.Element(("jabber:client", "message"))
        if to != '':
            message['to']   = to
        if type != '':
            message['type'] = typ
        if id != 0:
            message['id']   = m_id
            
        if body != '':
            message.addElement('body', content = body)
        if subject != '':
            message.addElement('subject', content = subject)
                        
        self.xmlstream.send(message)
        
        return message

    def setRoster(self, to, nick, group):
        """
        Send a roster set iq xmpp element.
        """
        self.touch()
        d = defer.Deferred()
        
        iq = client.IQ(self.xmlstream, "set")
        iq.addElement(("jabber:iq:roster", "query"))
        iq.query.addElement("item").attributes = {'jid' : to , 'name' : nick}
        try:
            if group:
                for g in group.split(' '):
                    iq.query.item.addElement('group', content = g)
        except:
            raise
        
        iq.addCallback(self._returnIq, d)
        iq.send()

        return d

    def disco(self, disco_jid, items = 1, auto=0):
        """
        Do the disco
        """
        self.touch()
        d = defer.Deferred()
        ret = []
        # TODO - this needs work
        if items == 1:
            stage = 1
        elif auto == 1:
            stage = 2
        else:
            stage = 0
        # TODO - disco.Disco needs to be simplified    
        iq = disco.Disco(self.xmlstream, "get")            
        if disco_jid != '':
            iq['to'] = disco_jid
        else:
            iq['to'] = self.jid.host

        iq.addElement((disco.NS_INFO, "query"))
        iq.addCallback(self._continue_disco, ret, stage, d)
        iq.send()
        return d
    

    # TODO - make this similar to disco, and fix this in xmlrpc.py
    def browse(self, b_jid, items=1,auto=0):
        """
        Do the browse
        """
        self.touch()
        d = defer.Deferred()
        self.b = []
        try:
            iq = disco.Disco(self.xmlstream, "get")            
            if b_jid != '':
                iq['to'] = b_jid
            else:
                iq['to'] = self.jid.host
            
            iq.addElement((disco.NS_INFO, "query"))
            iq.addCallback(self._returnBrowse, d)        
            iq.send()
            iq = disco.Disco(self.xmlstream, "get")
            if b_jid != '':
                iq['to'] = b_jid
            else:
                iq['to'] = self.jid.host
            
            iq.addElement((disco.NS_ITEMS, "query"))
            iq.addCallback(self._returnBrowse)
            iq.send()
            iq = client.IQ(self.xmlstream, "get")
            if b_jid != '':
                iq['to'] = b_jid
            else:
                iq['to'] = self.jid.host
            iq.addElement(("jabber:iq:browse", "query"))
            iq.addCallback(self._returnBrowse, d)
            iq.send()
        except:
            raise
        
        return d
        
    def _returnBrowse(self, iq, d = None):
        """
        An iq call back to return the iq request element.
        """
        self.touch()
        retval = []
        if len(self.b)<2:
            self.b.append(iq)
        else:
            self.b.append(iq)
            
            # loop through returned iqs
            for k, v in enumerate(self.b):
                if v.hasAttribute('type'):
                    if v['type'] == 'result':
                        if v.firstChildElement().name == "query":
                            ret = v.query.attributes
                            ret['items'] = []
                            if v.query.uri == 'http://jabber.org/protocol/disco#items':
                                items  = v.query.elements()
                                for i, item in enumerate(items):
                                    ret['items'].append(item.attributes)
                            elif v.query.uri == 'http://jabber.org/protocol/disco#info':
                                ret['identity'] = v.query.identity.attributes
                                ret['features'] = []
                                items  = v.query.elements()
                                for item in items:
                                    if item.name == 'feature':
                                        ret['features'].append(item.attributes)
                        elif v.firstChildElement().name == "service":
                            ret = v.service.attributes
                            ret['services'] = []
                            services  = v.service.elements()
                            for s in services:
                                ret['services'].append(s.attributes)
                        retval.append(ret)
            if d:        
                d.callback(retval)
                
        return 0
        
    def startRoster(self):
        """
        Request the roster
        """
        self.touch()
        d = defer.Deferred()
        
        iq = client.IQ(self.xmlstream, "get")
        iq.addElement(("jabber:iq:roster", "query"))
        iq.addCallback(self._returnRoster, d)
        iq.send()

        return d


    def _returnRoster(self, d, iq):
        """
        A iq call back to return the roster query result.
        """
        retval = []       
        items  = iq.query.elements()
        for i, item in enumerate(items):
            roster        = {}
            roster['jid'] = item['jid']
            
            if item.hasAttribute('name'):
                roster['name']   = item['name']
            if item.hasAttribute('ask'):
                roster['ask']    = item['ask']
            if item.hasAttribute('subscription'):
                roster['subscription']  = item['subscription']
                
            roster['id']     = i
            roster['groups'] = []
            try:
                groups         = item.elements()
                for i, group in enumerate(groups):
                    cgroup = u''	
                    for n in group.children:
                        if isinstance(n, types.StringTypes): cgroup = cgroup + n                    
                    roster['groups'].append(cgroup)
            except:
                pass
                # TODO - do something here?
            retval.append(roster)
        if d:
            d.callback(retval)
        else:
            return retval

    def vcard(self, jid):
        """
        Request a vcard
        """
        self.touch()
        d   = defer.Deferred()
        
        iq       = client.IQ(self.xmlstream, "get")
        iq['to'] = jid
        iq.addElement(("vcard-temp", "vCard"))
        iq.vCard['prodid']  = '-//HandGen//NONSGML vGen v1.0//EN'
        iq.vCard['version'] = '2.0'
        iq.addCallback(self._returnVcard, d)
        iq.send()

        return d


    def _returnVcard(self, d, iq):
        """
        A vcard callback to return the result from the server.
        """
        retval = {}
        if iq['type'] == 'error':
            if d:
                d.errback(error.Error(iq.error['code']))
            
        if iq.firstChildElement().name == 'vCard':
            items  = iq.vCard.elements()
            for i, item in enumerate(items):
                retval[item.name] = item
        if d:
            d.callback(retval)
            
        

    def _returnIq(self, d, iq):
        """
        An iq callback to return the result from the server.
        """
        try:
            t = (self, iq)
            if d:
                d.callback(t)
            else:
                self.elems.append(iq)
        except:
            raise
            # should we do something with this error?
            #pass
            

    def _continue_disco(self, retval = None, stage=0, iqd = None, d = None):
        """ """
        if not retval:
            retval = []
        # stage this, first grab info, then query for items?
        elems = iqd.query.children
        for e in elems:
            if not isinstance(e, types.StringTypes):
                retval.append({e.name : e.attributes})
        if stage == 0 and d:
            d.callback(retval)
            
        else:
            iq = disco.Disco(self.xmlstream, "get")            
            iq['to'] = iqd['from'] # this will change later
            iq.addElement((disco.NS_ITEMS, "query"))
            iq.addCallback(self._continue_disco, retval , stage-1)
            iq.send()


    def _returnNode(self, node, d = None):
        """ """
        try:
            if d:
               d.callback(node)
                
        except:
            # should we do something with this error?
            pass
            

    def getPresence(self, to, f):
        """ """
        self.touch()
        retval = []
        for i, p in enumerate(self.prs):
            presence = {}
            presence['time_stamp'] = 0
            for c in p.elements():
                if not isinstance(c, types.StringTypes) and not isinstance(c, types.NoneType):
                    cdata = u''	
		    for n in c.children:
			if isinstance(n, types.StringTypes):
				cdata = cdata + n

                    if c.name == 'priority':
                        presence['priority']   = cdata
                    if c.name == 'show':
                        presence['show']       = cdata
                    if c.name == 'status':
                        presence['status']     = cdata

            if p.hasAttribute('to'):
                presence['to_attr']    = p['to']
            if p.hasAttribute('from'):
                presence['from_attr']  = p['from']
            if p.hasAttribute('id'):
                presence['id']    = p['id']
            else:
                presence['id']    = i
            
            if p.hasAttribute('type'):
                presence['type']    = p['type']
            
            presence['xml']        = p.toXml()
            if to != '' and p['to'].find(to) != -1:
                retval.append(presence)
            elif f != '' and p['from'].find(f) != -1:
                retval.append(presence)    
            elif to == '' and f == '':
                retval.append(presence)    

        return retval

    def getMessage(self, to, f, type, id):
        """ """
        return self.checkMessage(to, f, type, id, True)


    def checkMessage(self, to, f, type, id, delete = False):
        """ """
        self.touch()
        retval = []
        d = []
        for i, m in enumerate(self.msgs):
            message = {}            
            
            message['time_stamp'] = 0
            for c in m.elements():
                if not isinstance(c, types.StringTypes) and not isinstance(c, types.NoneType):
                    cdata = u''	
		    for n in c.children:
			if isinstance(n, types.StringTypes):
				cdata = cdata + n
                    
                    if c.name == 'subject':
                        message['subject']    = cdata
                    if c.name == 'body':
                        message['body']       = cdata
                    if c.name == 'status':
                        message['status']     = cdata


            if m.hasAttribute('to'):
                message['to_attr']    = m['to']
            if m.hasAttribute('from'):
                message['from_attr']  = m['from']
            if m.hasAttribute('thread'):
                message['thread']     = m['thread']
            if m.hasAttribute('id'):
                message['id']    = m['id']
            else:
                message['id']    = i
            
            if m.hasAttribute('type'):
                message['type']    = m['type']
            
            message['xml']        = m.toXml()

            
            if len(to)>0 and m['from'].find(to) != -1:
                retval.append(message)
                if delete:
                    d.append(m)
            elif len(f)>0 and m['to'].find(f) != -1:
                retval.append(message)
                if delete:
                    d.append(m)
            elif len(type)>0 and m['type'].find(type) != -1:
                retval.append(message)
                if delete:
                    d.append(m)
            elif m.hasAttribute('id'):
                if id == 0 or m['id'] == id:
                    retval.append(message)
                    if delete:
                        d.append(m)
            elif len(to)==0 and len(f)==0 and len(type)==0 and id == 0:
                retval.append(message)    
                if delete:
                    d.append(m)
        if delete:
            for m in d:
                if m in self.msgs:
                    del self.msgs[self.msgs.index(m)]
        return retval


    def joinChat(self, client, server, nick, room, mcb, pcb):
        """
        @client - this is the call back class of cb
        
        """
        self.chat_p = self.chat_p + 1
        if self.group_chats.has_key(room+'@'+server):
            return
        # TODO - fix this
        #message_event  = "/message[@from='"+room+"@"+server+"']"
        #presence_event = "/presence[@from='"+room+"@"+server+"']"
        message_event  = "/message[@type='groupchat']"
        presence_event = "/presence/x"
        
        try:
            self.xmlstream.addObserver(message_event, mcb, self.chat_p, client, server, nick, room) # What do we do with priority?
            self.xmlstream.addObserver(presence_event, pcb, self.chat_p, client, server, nick, room) # What do we do with priority?
        except:
            raise
        presence = domish.Element(("jabber:client","presence"))
        presence['to'] = room + '@'+ server + '/' + nick
        presence.addElement(('http://jabber.org/protocol/muc','x'))
        reactor.callLater(1, self.xmlstream.send, presence)
        

    def partChat(self, room, mcb, pcb):
        """ Leave a group chat """
        #message_event  = "/message[juserhost(@from)='"+room+"']"
        #presence_event = "/presence[juserhost(@from)='"+room+"']"
        message_event  = "/message[@type='groupchat']"
        presence_event = "/presence/x"
        presence = domish.Element(("jabber:client","presence"))
        presence['to']   = room
        presence['type'] = 'unavailable'
        presence.addElement(('http://jabber.org/protocol/muc','x'))
        self.xmlstream.send(presence)

        try:
            self.xmlstream.removeObserver(message_event, mcb)
            self.xmlstream.removeObserver(presence_event, pcb)
        except:
            raise

        
    def getIq(self, iqid):
        """ """
        self.touch()
        retval   = []
        if iqid != 0:
            for i, iq in enumerate(self.iqs):
                if iq.hasAttribute('id'):
                    if iqid == iq['id']:
                        return {iqid : iq}
        else:
            retval = self.iqs
            self.iqs = []
        return retval
        
    def iqHandler(self, xs, priority = 0):
        """ Handle and store iq xmpp elements """
        try:
            self.iqs.append(xs)
        except:
            return 0


    def preHandler(self, xs, priority = 0):
        """ """
        f = None
        try:
            if xs.hasAttribute('from'):
                f = xs['from']
            # is there another way to loop this better?
            for k, v in enumerate(self.prs):
                if v.hasAttribute('from') and f:
                    if v['from'].find(f) != -1:
                        del self.prs[k]
                        
            self.prs.append(xs)
        except:
            return 0


    def msgHandler(self, msg, priority = 0):
        """ """
        try:
            self.msgs.append(msg)
        except:
            return 0

    
    def checkExpired(self):
        """
        Check if the session or xmpp connection has expired
        """
        # send this so we do not timeout from servers
        if getattr(self, 'xmlsteam', None):
            self.xmlstream.send(' ')
        if self.inactivity is None:
            wait = 900
        elif self.inactivity == 0:
            wait = time.time()
        
        else:
            wait = self.inactivity

        if self.waiting_requests and len(self.waiting_requests)>0:
            wait += self.wait # if we have pending requests we need to add the wait time
            
        if time.time() - self.lastModified > wait+(0.1):
            if self.site.sessions.has_key(self.uid):
                self.terminate()
            else:
                pass
             
        else:
            reactor.callLater(wait, self.checkExpired)


    def _cacheData(self, rid, data):
        if len(self.cache_data.keys())>=3:
            self.cache_data = {}
        self.cache_data[rid] = data
        
# This stuff will leave when SASL and TLS are implemented correctly
# session stuff

    def _sessionResultEvent(self, iq):
        """ """
	if len(self.waiting_requests)>0:		
		wr = self.waiting_requests.pop(0)
		d  = wr.deferred
	else:
		d = None

        if iq["type"] == "result":
            if d:
                d.callback(self)
        else:
            if d:
                d.errback(self)

    # BIND stuff
    def _bindResultEvent(self, iq):
        """ """
	if iq["type"] == "result":
            self.bind = 1
                
        else:
            self.bind = 0
            
	if len(self.waiting_requests)>0:		
		wr = self.waiting_requests.pop(0)
		d = wr.deferred
	else:
		d = None
        
	if self.bind == 1 and self.session == 1:
		if d:
	            d.callback(self)
        else:
		if d:
	            d.errback(self)
            
    # SASL stuff
    

    def _saslError(self, sasl_error, d = None):
        """ SASL error """
        log.msg('\n\nSASL ERROR\n\n')
        if d:
            d.errback(self)
        if len(self.waiting_requests)>0:		
            wr = self.waiting_requests.pop(0)
            wr.doCallback([sasl_error])
            
    def _saslStep1(self, challenge):
        """ """
        c = str(challenge)
        
        dc = base64.decodestring(c)
        ra = _parse(dc)
        if ra.has_key('realm'):
            self.realm = ra['realm']
        else:
            self.realm = self.jid.host
        self.nonce = ra['nonce']
        self.nc = 0
        self.charset = ra['charset']
        self.algorithm = ra['algorithm']
        response = domish.Element((client.SASL_XMLNS, "response"))

        r = self._response(self.charset, self.realm, self.nonce)
        
        response.addContent(r)
        self.xmlstream.removeObserver("/challenge", self._saslStep1)
        self.xmlstream.addOnetimeObserver("/challenge", self._saslStep2)
        self.xmlstream.send(response)


    def _saslStep2(self, challenge):
        """ """
        cs = base64.decodestring(str(challenge))
        ca = _parse(cs)
        
        if self.rauth == ca['rspauth']:
            response = domish.Element((client.SASL_XMLNS, "response"))
                        
            self.xmlstream.removeObserver("/challenge", self._saslStep2)
            self.xmlstream.addOnetimeObserver("/success", self._saslSuccess)
            self.xmlstream.send(response)
        else:
            self.xmlstream.dispatch(challenge, self.AUTH_FAILED_EVENT)


    def _saslSuccess(self, s):
        """ """
        self.success = 1
        self.s = s
        # return success to the client
        if len(self.waiting_requests)>0:		
            wr = self.waiting_requests.pop(0)
            wr.doCallback([s])
        self.authenticator._reset()
        from httpb  import HttpbService
        from httpp  import HttppService
        if not isinstance(self.pint, HttpbService) and not isinstance(self.pint, HttppService) :
            log.msg('==================================== Add New Observers not for BOSH ==================')
            self.xmlstream.addObserver("/message",  self.msgHandler, -1)
            self.xmlstream.addObserver("/presence", self.preHandler, -1)
                

    # SASL stuff - maybe put this in its own class?
    def _response(self, charset, realm, nonce):
        """ """
        rs = ''
        try:
            username = self.jid.user.encode(charset)
        except UnicodeError:
            # TODO - add error checking 
            raise
        rs = rs + 'username="%s"' % username
        rs = rs + ',realm="%s"' % realm
        cnonce = _gen_nonce()
        rs = rs + ',cnonce="%s"' % cnonce
        rs = rs + ',nonce="%s"' % nonce

        self.nc += 1
        nc = "%08x" % self.nc
        rs = rs + ',nc=%s' % nc
        rs = rs + ',qop=auth'

        rs = rs + ',digest-uri="xmpp/'+self.jid.host+'"'
        

        uh = "%s:%s:%s" % (username,
                           realm,
                           self.passwd.encode(self.charset))
        huh = md5.new(uh).digest()
        # TODO - add authzid
        a1 = "%s:%s:%s" % (huh, nonce, cnonce)
        a2 = "AUTHENTICATE:xmpp/"+self.jid.host

        a3 =  ":xmpp/"+self.jid.host

        resp1 = "%s:%s:%s:%s:%s:%s" % (binascii.b2a_hex(md5.new(a1).digest()),
                                       nonce,
                                       nc,
                                       cnonce,
                                       "auth",
                                       binascii.b2a_hex(md5.new(a2).digest()))
        
        resp2 = "%s:%s:%s:%s:%s:%s" % (binascii.b2a_hex(md5.new(a1).digest()),
                                       nonce,
                                       nc,
                                       cnonce,
                                       "auth",
                                       binascii.b2a_hex(md5.new(a3).digest()))
        
        kda1 = md5.new(resp1).digest()
        kda2 = md5.new(resp2).digest()
        
        response = binascii.b2a_hex(kda1) 
        
        rs = rs + ',response="%s"' % response
        rs = rs + ',charset=%s' % charset
        
        self.rauth = binascii.b2a_hex(kda2)
        
        return  binascii.b2a_base64(rs)
        
def _parse(self, rcs):
    """ Parse a SASL challenge """
    r = rcs.split(',')
    h = {}
    for i in r:
        (k, v) = i.split('=')
        v = v.replace("'","")
        v = v.replace('"','')
        if h.has_key(k):
            # return an error
            return 0
        h[k] = v
    return h



