# XMPP server class 

from twisted.application import internet, service
from twisted.internet import interfaces,protocol
from twisted.python import usage,components

# tmp
from twisted.protocols import basic
from twisted.internet import reactor, defer

try:
    from twisted.words.xish import domish, utility , xmlstream
except:
    from twisted.xish import domish, utility , xmlstream

import sha, md5, random,  types, os, time
from binascii import b2a_base64, a2b_base64, b2a_hex, a2b_hex

SASL_XMLNS = 'urn:ietf:params:xml:ns:xmpp-sasl'
COMP_XMLNS = 'http://jabberd.jabberstudio.org/ns/component/1.0'
STREAMS_XMLNS  = 'urn:ietf:params:xml:ns:xmpp-streams'

class XMPPServerComponentAuthenticator(xmlstream.Authenticator):
    
    def __init__(self,config=None):
        self.counter  = 1
        self.sessions = {}
        self.config   = config
        self.step     = 0
    
    def connectionMade(self):
        if self.config.has_key('debug') and self.config['debug'] >0:
            print "Connection Made"
        

    def notWellFormedError(self, obj):
        self.sendError('xml-not-well-formed','syntax error')
        
    def sendError(self, error, text = None):
        # TODO - add version stuff!
        # TODO - make this an domish element?
        if self.counter == 1:
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0'><stream:error xmlns:stream='http://etherx.jabber.org/streams'>"
        else:
            sh = "<stream:error xmlns:stream='http://etherx.jabber.org/streams'>"
        # TODO - check for valid error types, create a map
        sh = sh + "<%s xmlns='%s' />" % (error, STREAMS_XMLNS)
        if text:
            sh = sh + "<text xmlns='%s' >%s</text>" % (STREAMS_XMLNS, text)

        if self.counter == 1:
            sh = sh + "</stream:error></stream:stream>"    
        else:
            sh = sh + "</stream:error>"    

        self.xmlstream.send(sh)
            
    def streamStarted(self ,xs):
        # CHECK FOR ERROR
        # TODO - add version stuff!
        if self.step == 2:
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0' id='%s'>" % \
                 (self.sid)
            self.xmlstream.send(sh)
            # send features
            # TODO - loop through the REAL stuff
            sh = """<stream:features>
  <bind xmlns='urn:ietf:params:xml:ns:xmpp-bind'/>
</stream:features>"""
            # TODO - add an event here. this is bad
            
            self.xmlstream.addOnetimeObserver("/bind", self.startBind)
            
        else:
            # Generate stream header
            self.counter = self.counter + 1
            self.sid  = md5.new("%s_%s" % (str(random.random()) , str(self.counter))).hexdigest()
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0' id='%s'>" % \
                 (self.sid)
            self.xmlstream.send(sh)
            # send features
            # TODO - loop through the REAL stuff
            # TODO - add tls support <starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>
            sh = """<stream:features>
            
            <mechanisms xmlns='urn:ietf:params:xml:ns:xmpp-sasl'>
            <mechanism>DIGEST-MD5</mechanism>
            <mechanism>PLAIN</mechanism>
            </mechanisms>
            </stream:features>"""
            # TODO - handle old jabber component connections
            self.xmlstream.addOnetimeObserver("/starttls", self.startTLS)
            self.xmlstream.addOnetimeObserver("/auth", self.startSASL)

        self.xmlstream.send(sh)
            

    def startSASL(self, elem):
        self.step = 1
        challenge = domish.Element((SASL_XMLNS,'challenge'))
        challenge['xmlns'] = SASL_XMLNS # why does the above not work?
        if elem.hasAttribute('mechanism') and elem['mechanism'] == "DIGEST-MD5":
            nonce = self._gen_nonce()
            self.nonce = nonce
            params = []

            #cs = 'realm="%s",nonce="%s",qop="auth",charset=utf-8,algorithm=md5-sess' % (self.config['id'], nonce)
            params.append('realm="%s"' % (self.config['id'],))
            params.append('nonce="%s"' % (nonce,))
            params.append('qop="auth"')
            params.append('charset=utf-8')
            params.append('algorithm=md5-sess')
            #bcs = base64.encodestring(cs)
            bcs=b2a_base64(",".join(params))
            if bcs[-1]=='\n':
                bcs=bcs[:-1]
            
            challenge.addContent(bcs)
                                 
        
        elif elem.hasAttribute('mechanism') and elem['mechanism'] == "PLAIN":
            print "PLAIN"
        else:
            print "Send Failure"

        self.xmlstream.addOnetimeObserver("/response", self._saslStep2)
        self.xmlstream.send(challenge)

    def _saslStep2(self, elem):
        if self.step == 1:
            self.step = 2
        else:
            return self._sasl_error('temporary-auth-failure')
        
        #response = base64.decodestring(str(elem))
        response = a2b_base64(str(elem))
        
        r = response.split(',')
        h = {}
        for i in r:
            (k,v) = i.split('=')
            v = v.replace("'","")
            v = v.replace('"','')
            if h.has_key(k):
                # return an error
                return self._sasl_error('temporary-auth-failure')   
            h[k] = v

        if (h.has_key('username') and h.has_key('realm')
            and h.has_key('nonce') and h.has_key('cnonce')
            and h.has_key('response') and  h.has_key('charset')
            and h.has_key('nc')):
            passwd = ''
            if h.has_key('username') and self.config['local']['users'].has_key(h['username']):
                self.user = h['username'].encode('utf-8')
                passwd = self.config['local']['users'][h['username']]
            else:
                self.user = None
                passwd = self.config['local']['secret']
            passwd = passwd.encode("utf-8")
            
            sp  = ":".join([h['username'].encode("utf-8"),h['realm'].encode("utf-8"),passwd])
            
            msp = md5.new(sp).digest()
            
            a1 = ":".join([msp,h['nonce'],h['cnonce']])
            if h.has_key('authzid'):
                print "authzid " + h['authzid']
                a1 = ":".join(a1,h['authzid'])
                        
            a2 = "AUTHENTICATE:"+h['digest-uri']
            
            ha1 = b2a_hex(md5.new(a1).digest())
            ha2 = b2a_hex(md5.new(a2).digest())
            check = b2a_hex(md5.new(":".join(
                [ha1,
                 h['nonce'],
                 h['nc'],
                 h['cnonce'],
                 h['qop'],
                 ha2])).digest())
            
            
            if check != h['response']:
                return self._sasl_error('temporary-auth-failure')
            else:
                rspauth = md5.new(":".join([md5.new(a1).hexdigest(),
                                            h['nonce'],
                                            h['nc'],
                                            h['cnonce'],
                                            h['qop'],
                                            md5.new(":"+h['digest-uri']).hexdigest()
                                            ])
                                  ).hexdigest()
                rsvp = 'rspauth="'+rspauth+'"';
                challenge = domish.Element((SASL_XMLNS,'challenge'))
                challenge['xmlns'] = SASL_XMLNS # why does the above not work?
                challenge.addContent(b2a_base64(rsvp))
                self.xmlstream.removeObserver("/response", self._saslStep2) # addOnetime is not working for me right now :(
                self.xmlstream.addOnetimeObserver("/response", self._saslStep3)
                self.xmlstream.send(challenge)
        else:
            return self._sasl_error('temporary-auth-failure')

    def _saslStep3(self,elem):
        if self.step == 2:
            #print "SASL STEP 3"
            self.stage = 3
        else:
            #print "WRONG SASL STEP"
            return self._sasl_error('temporary-auth-failure')

        success = domish.Element((SASL_XMLNS,'success'))
        success['xmlns'] = SASL_XMLNS
        self.xmlstream.removeObserver("/response", self._saslStep3) # addOnetime is not working for me right now :(
        #self.xmlstream.addOnetimeObserver("/response", self._saslStep3)
        self.xmlstream.send(success)
        # RESET ?
        self.xmlstream._reset()
        
        
        
    def _gen_nonce(self):
        return md5.new("%s:%s:%s" % (str(random.random()) , str(time.gmtime()),str(os.getpid()))).hexdigest()
            
    def _sasl_error(self,error):
        fail = domish.Element((SASL_XMLNS,'failure'))
        fail['xmlns'] = SASL_XMLNS
        fail.addElement(error)
        self.xmlstream.send(fail)
        return

    def startTLS(self, elem):
        print "Start TLS"
        print self.sid
        print self.counter


class XMPPServerClientAuthenticator(XMPPServerComponentAuthenticator):
    
    def __init__(self,config=None):
        self.counter  = 1
        self.sessions = {}
        self.config   = config
        self.step     = 0
    
    def connectionMade(self):
        if self.config.has_key('debug') and self.config['debug'] >0:
            print "Connection Made"
        
        
    def notWellFormedError(self, obj):
        self.sendError('xml-not-well-formed','syntax error')
        
    def sendError(self, error, text = None):
        # TODO - add version stuff!
        # TODO - make this an domish element?
        if self.counter == 1:
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0'><stream:error xmlns:stream='http://etherx.jabber.org/streams'>"
        else:
            sh = "<stream:error xmlns:stream='http://etherx.jabber.org/streams'>"
        # TODO - check for valid error types, create a map
        sh = sh + "<%s xmlns='%s' />" % (error, STREAMS_XMLNS)
        if text:
            sh = sh + "<text xmlns='%s' >%s</text>" % (STREAMS_XMLNS, text)

        if self.counter == 1:
            sh = sh + "</stream:error></stream:stream>"    
        else:
            sh = sh + "</stream:error>"    

        self.xmlstream.send(sh)


    def startBind(self,bind):
        print bind.toXml().encode('utf-8')

    def startSession(self,session):
        print session.toXml().encode('utf-8')

    def handleIQ(self, iq):
        print iq.toXml()

    def startTLS(self, elem):
        print "Start TLS"
        print self.sid
        print self.counter

        

    def streamStarted(self ,xs):
        # CHECK FOR ERROR
        
        if self.step == 2:
            print "\n\n\n\n" + str(self.step) + "\n\n\n\n"
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0' id='%s'>" % \
                 (self.sid)
            self.xmlstream.send(sh)
            # send features
            # TODO - loop through the REAL stuff
            sh = """<stream:features>
  <bind xmlns='urn:ietf:params:xml:ns:xmpp-bind'/>
  <session xmlns='urn:ietf:params:xml:ns:xmpp-session'/>
</stream:features>"""
            # TODO - this is bad
            self.xmlstream.addOnetimeObserver("/bind", self.startBind)
            self.xmlstream.addOnetimeObserver("/session", self.startSession)
            
        else:
            # Generate stream header
            self.counter = self.counter + 1
            self.sid  = md5.new("%s_%s" % (str(random.random()) , str(self.counter))).hexdigest()
            sh = "<stream:stream xmlns:stream='http://etherx.jabber.org/streams' version='1.0' id='%s'>" % \
                 (self.sid)
            self.xmlstream.send(sh)
            # send features
            # TODO - loop through the REAL stuff
            sh = """<stream:features>
            <starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>
            <mechanisms xmlns='urn:ietf:params:xml:ns:xmpp-sasl'>
            <mechanism>DIGEST-MD5</mechanism>
            <mechanism>PLAIN</mechanism>
            </mechanisms>
            </stream:features>"""
            
            self.xmlstream.addOnetimeObserver("/starttls", self.startTLS)
            self.xmlstream.addOnetimeObserver("/auth", self.startSASL)
            self.xmlstream.addOnetimeObserver("/iq", self.handleIQ)

        self.xmlstream.send(sh)
